import unittest
from khanaa import spelling_decompose

GENERAL = {
    'เขียน': {
        'data': {
            'onset': 'ข',
            'vowel': 'เอีย',
            'silent_before': '',
            'coda': 'น',
            'silent_after': '',
            'tone': 4
            },
        'detail': {
            'tone_mark': '',
            'leading_h': False,
            'vowel_form': 'เ-ี+ย',
            'onset_index': -1,
            'onset_main': 'ข'
            },
        'pref': {}
        },
    'สมุทร': {
        'data': {
            'onset': 'สม',
            'vowel': 'อุ',
            'silent_before': '',
            'coda': 'ท',
            'silent_after': 'ร',
            'tone': 1
            },
        'detail': {
            'tone_mark': '',
            'leading_h': False,
            'vowel_form': '-ุ+',
            'onset_index': -2,
            'onset_main': 'ส'
            },
        'pref': {
            'silent_after_style': 'plain',
            'clear_vowel': False,
            'clear_vowel_onset': 'not_true_cluster',
            'clear_vowel_tone_mark': False
            }
        },
    'แถลง': {
        'data': {
            'onset': 'ถล',
            'vowel': 'แอ',
            'silent_before': '',
            'coda': 'ง',
            'silent_after': '',
            'tone': 4
            },
        'detail': {
            'tone_mark': '',
            'leading_h': False,
            'vowel_form': 'แ-+',
            'onset_index': -2,
            'onset_main': 'ถ'
            },
        'pref': {
            'clear_vowel': False,
            'clear_vowel_onset': 'not_true_cluster',
            'clear_vowel_tone_mark': False
            }
        }
}

class TestSpellingDecompose(unittest.TestCase):

    def test_general(self):
        for case in GENERAL:
            result = spelling_decompose(case)
            self.assertEqual(result, GENERAL[case])

if __name__ == '__main__':
    unittest.main()


"""This module contains class that combine word data with methods."""

from typing import List

from khanaa.ambiguity import find_donee_end, find_donor_end, find_donor_start
from khanaa.homophone import find_homophone_product
from khanaa.pronunciation import ThaiToIPA
from khanaa.speller import combine
from khanaa.word import Word

class Combination(Word):
    """Combine finding word data part from Word with methods.
    
    For information for each methods, see Kham.
    """
    
    @property
    def form(self) -> str:
        """Return spelled form of the word."""
        return combine(self, self.pref)
    
    @property
    def all_tone(self) -> List[str]:
        all_tone = []
        for n in range(5):
            spell = Combination(self._onset, self._vowel, self._silent_before,
                self._coda, self._silent_after, n, **self.pref)
            if not spell._is_possible_tone:
                all_tone.append('')
                continue
            all_tone.append(spell.form)
        return all_tone

    @property
    def reading(self) -> str:
        # it's _tone_realized not _tone
        # because we want spelled word sound, not the input one.
        thai_to_ipa = ThaiToIPA(self._onset, self._vowel, self._silent_before,
            self._coda, self._silent_after, self._tone_realized, **self.pref)
        return thai_to_ipa.convert()
    
    @property
    def is_donee_end(self) -> bool:
        return find_donee_end(self._vowel, self._silent_before, self._coda,
            self._silent_after, self.pref['silent_before_style'],
            self.pref['silent_after_style'])
    
    @property
    def is_donor_end(self) -> bool:
        return find_donor_end(self._vowel, self._silent_before, self._coda,
            self._silent_after, self.pref['silent_before_style'],
            self.pref['silent_after_style'], self._tone_mark,
            self._vowel_form)
    
    @property
    def is_donor_start(self) -> bool:
        return find_donor_start(self._onset, self.form)
    
    @property
    def homophone(self) -> list:
        product = find_homophone_product(self._onset, self._vowel, self._coda)
        result = []
        for possible in product:
            kham = Combination(
                onset=possible[0],
                vowel=possible[1],
                coda=possible[2],
                tone=self._tone_realized
            )
            result.append(kham.form)
        return list(dict.fromkeys(result))


"""This module stores function that decompose word's components.

This function uses basic rule to decompose syllable,
so it won't cover some cases:
- เทอม (เอ) No เออ as เออ coda form yet
- ก็ will be None. (Don't have อ็ vowel in dict yet)
- มารถ will be มา+ร+ถ์
- ฤทธิ์ will have โอะ vowel
- Overall ฤ won't have correct sound ex. ทฤษ(ฎี)
- โพล will be พ+โอ+ล although it can be both this and พล+โอ when
judged from the writing
- Some short vowel use ไม้ไต่คู้ when it has coda, but when it also
has tone mark, ไม้ไต่คู้ will be delete so its form will look like
its long vowel pair.
- Some stylistic data are not saved such as if silent_before doesn't
have kaaran, so if we use the data with Kham, it might have
different form.
- Not covering phinthu, yamakkaan
"""

import re
from typing import Any, Dict, List, Tuple, Union

from khanaa.thai_script import (CLUSTERS, CONSONANTS, DIACRITICS,
    TONE_MARKERS, VOWEL_CHAR, VOWELS)
from khanaa.word import Word
from khanaa.utils import find_tone

TONE_MARK: str = ''.join([TONE_MARKERS[tone] for tone in TONE_MARKERS])
CODA_LIST: str = ''.join([char for char in CONSONANTS
    if CONSONANTS[char]['sound_coda']
    and char not in ['ย', 'ว']])

vowel_re = None
# should be created once by create_vowel_re()
# and after that its value won't be changed

def spelling_decompose(text: str) -> Union[Dict[str, Any], None]:
    """Find each part of the spelled word.

    Args:
        text: A Thai syllable such as เขียน (Also fine in case of
            multiple onsets such as มหา, แถลง, สมุทร)

    Returns:
        Syllable data in this structure:
            - 'data': onset, vowel, silent_before, coda, silent_after,
            tone
            - 'detail': tone_mark, leading_h, vowel_form, onset_index,
            onset_main
            - 'pref': (in some case)
                - vowel_coda_form
                - low_single_h_thoo
                - clear_vowel
                - clear_vowel_onset
                - clear_vowel_tone_mark
                - silent_after_style
            
        Or None if it can't be analyzed.
    """
    if not text:
        return

    global vowel_re
    if vowel_re == None:
        vowel_re = create_vowel_re()
    pref = {}

    tone_mark = analyze_tone_mark(text)
    no_sa, silent_after_last = split_silent_after(text)
    # no_sa is from no_silent_after

    vowel, silent_after, no_sa, vowel_type, vowel_form = analyze_vowel(no_sa)
    if not vowel:
        return

    if silent_after and not silent_after_last:
        pref.update({'silent_after_style': 'plain'})
    silent_after = ''.join([silent_after, silent_after_last])
    # remove tone mark after find vowel
    # because sometimes tone mark can help disambiguate the word
    # ex. สวน=ส+อัว+น or สว+โอะ+น? but ส่วน is only ส+อัว+น
    # but we don't want it while finding onset and coda
    no_sa = remove_tone(no_sa)
    
    coda, no_coda = analyze_coda(no_sa, vowel_type)
    onset, leading_h, onset_f, onset_b = analyze_onset(no_coda, vowel_form)
    # onset_f is onset char(s) that comes before front vowel such as
    # ส in สเตย์ while onset_b is onset char(s) that comes after it
    # such as ต
    if len(onset_f) + len(onset_b) == 2:
        pref.update(find_onset_pref(onset_f, onset_b, tone_mark))
    silent_before = analyze_silent_before(no_coda)

    # โอะ+ร > ออ+ร
    new_vowel, new_form = change_o_r(vowel, coda)
    if new_form:
        vowel = new_vowel
        vowel_form = new_form
        pref.update({'vowel_coda_form': {new_vowel: new_form}})

    # find tone number
    is_low_single_vague = Word._find_is_low_single_vague(True, onset)
    is_h_vague = Word._find_is_h_vague(True, onset)
    onset_index = Word._find_onset_index(is_low_single_vague,
        is_h_vague, onset, False, False, False)
    onset_main = Word._find_onset_main(onset, onset_index)
    tone = find_tone(onset_main, vowel, coda, tone_mark, leading_h)

    # ห+ต่ำเดี่ยว+ไม้โท
    if save_h_thoo(tone, leading_h):
        pref.update({'low_single_h_thoo': True})

    return {'data': {'onset': onset, 'vowel': vowel,
            'silent_before': silent_before, 'coda': coda,
            'silent_after': silent_after, 'tone': tone},
        'detail': {'tone_mark': tone_mark, 'leading_h': leading_h,
            'vowel_form': vowel_form, 'onset_index': onset_index,
            'onset_main': onset_main},
        'pref': pref}

def replace_vowel_re(form: str) -> str:
    """Add regular expression to vowel form.

    Args:
        form: Vowel form

    Returns:
        Regular expression string for the vowel form.
    """
    form: str = form.replace('+', f'[{TONE_MARK}]?')
    form = form.replace('-', '[ก-ฮ]+')
    form = ''.join(['^[ก-ฮ]*', form])
    return form

def create_vowel_re():
    """Create vowel regular expression dict to be used.

    Returns:
        Vowel dict in this structure:
            - 'vowel_jw' for vowel with ย, ว coda
            - 'vowel_coda' for vowel with coda
            - 'vowel_no_coda' for vowel without coda
            - 'ex_coda' for vowel with coda that should come after
            vowel_no_coda
            - 'ex_no_coda' for vowel without coda that should come
            after ex_no_coda
            
        Each vowel type contains vowels sorted by vowel form length
        and each vowel contains:
            - 'form' for vowel form
            - 're' for vowel regex
    """
    # find vowels for each vowel type because each vowel type will use
    # different regex
    vowel_re = {'vowel_jw': {}, 'vowel_coda': {}, 'vowel_no_coda': {}}
    for vowel in VOWELS:
        if (VOWELS[vowel]['form_no_coda']
                and VOWELS[vowel]['sound_coda'] in ['j', 'w']):
            vowel_re['vowel_jw'].update({vowel: VOWELS[vowel]['form_no_coda']})
        if VOWELS[vowel]['form_with_coda']:
            vowel_re['vowel_coda'].update({vowel: VOWELS[vowel]['form_with_coda']})
        if (VOWELS[vowel]['form_no_coda']
                and VOWELS[vowel]['sound_coda'] not in ['j', 'w']):
            vowel_re['vowel_no_coda'].update({vowel: VOWELS[vowel]['form_no_coda']})
    
    # these vowels should come last because they're invisible
    vowel_re['vowel_coda'].pop('อ')
    vowel_re['vowel_coda'].pop('โอะ')
    vowel_re['vowel_no_coda'].pop('อ')
    vowel_re.update({'ex_coda': {'โอะ': VOWELS['โอะ']['form_with_coda']}})
    vowel_re.update({'ex_no_coda': {'อ': VOWELS['อ']['form_no_coda']}})

    # sort vowels and create regex for each vowel
    for vowel_type, vowel_data in vowel_re.items():
        sorted_vowel = sorted(vowel_re[vowel_type],
            key=lambda vowel: len(vowel_re[vowel_type][vowel]), reverse=True)
        additional = ''
        if vowel_type == 'vowel_jw':
            additional = f'(?![์{TONE_MARK}])'
        elif vowel_type in ['vowel_coda', 'ex_coda']:
            additional = f'([ก-ฮ]+์)?[{CODA_LIST}](?![์{TONE_MARK}])'
        new_vowel_data = {}
        for vowel in sorted_vowel:
            pattern = re.compile(
                replace_vowel_re(vowel_data[vowel]) + additional)
            new_vowel_data.update({
                vowel: {
                    'form': vowel_data[vowel],
                    're': pattern}})
        vowel_re[vowel_type] = new_vowel_data
    return vowel_re

def split_silent_after(text: str) -> Tuple[str, str]:
    """Split silent after from the last position.

    This cannot find more than one silent_after char, so if the word
    has more than one, we'll catch it when we find the vowel.
    Because sometimes silent_after range can be ambiguous.
    (สเตย์=สต+เอ+ย์ we shouldn't include ต because เอ char but
    จันทร์ we should include ท but not น because อั char needs coda).

    As for why we don't find all silent_after only after find vowel,
    because this way we can still catch words like ฤทธิ์
    which has silent vowel char under silent mark.

    Args:
        text: Thai syllable

    Returns:
        text without silent_after, silent_after
    """
    no_silent_after: str = text
    silent_after: str = ''
    if text[-1] == DIACRITICS['kaaran']:
        result: List[str] = re.split(f'([ก-ฮ][{VOWEL_CHAR}]?์$)', text)
        if len(result) > 1:
            no_silent_after = result[0]
            silent_after = result[1].replace('์', '')
    return no_silent_after, silent_after

def analyze_tone_mark(text: str):
    """Find tone mark used.

    Args:
        text: The word

    Returns:
        Tone mark name
    """
    tone_mark = ''
    for char in TONE_MARK:
        if char in text:
            tone_mark = char
            break
    # because tone marks are difficult to see
    # we'll return their names instead
    for name, tone_char in TONE_MARKERS.items():
        if tone_mark == tone_char:
            tone_mark = name
            break
    return tone_mark

def analyze_vowel(text: str) -> Tuple[str, str, str, str, str]:
    """Find vowel in the word.

    Args:
        text: Thai syllable.

    Returns:
        vowel, silent_after, word without silent_after, vowel_type
    """
    all_vowels: List[str] = [char for char in text if char in VOWEL_CHAR]
    vowel, silent_after, no_silent_after, vowel_form = '', '', '', ''
    for vowel_type in vowel_re:
        result = analyze_vowel_form(text, all_vowels, vowel_type)
        if result[0]:
            vowel, silent_after, no_silent_after, vowel_form = result
            break
    return vowel, silent_after, no_silent_after, vowel_type, vowel_form

def analyze_vowel_form(text: str, all_vowels: List[str],
        vowel_type: str) -> Tuple[str, str, str, str]:
    """Find vowel in the word from one vowel type.

    Args:
        text: One Thai syllable/word
        vowel_type: From vowel_re

    Returns:
        Vowel form if vowel is found. Empty string if not found.
    """
    for vowel in vowel_re[vowel_type]:
        form = vowel_re[vowel_type][vowel]['form']
        pattern = vowel_re[vowel_type][vowel]['re']
        result = re.search(pattern, text)
        if result:
            # To make sure every vowel char is taken into account.
            # Even if we order our search by vowel form length,
            # because we search vowel with ย, ว before other type
            # of vowel, the word like เปลี่ยน will be matched with
            # อีย before เอีย.
            leftover = [char for char in all_vowels if char not in form]
            if leftover:
                # for ญาติ, เหตุ
                if (len(leftover) == 1
                        and leftover[0] in ['ิ', 'ุ']
                        and text[-1] in ['ิ', 'ุ']):
                    pass
                else:
                    continue
            split = re.split(pattern, text)
            silent_after = ''
            if split[-1]:
                # แสวง should be สว+แอ+ง not ส+แอว+(ง)
                if (vowel_type == 'vowel_jw'
                        and split[-1].find(DIACRITICS['kaaran']) == -1):
                    continue
                silent_after = split[-1].replace(DIACRITICS['kaaran'], '')
            return vowel, silent_after, result[0], form
    return '', '', '', ''

def analyze_coda(text: str, vowel_type: str) -> Tuple[str, str]:
    """Find coda. This assumes that there's no silent_after left.

    Args:
        text: Syllable without silent_after
        vowel_type: From vowel_re

    Returns:
        coda, text without coda
    """
    coda = ''
    if vowel_type in ['vowel_coda', 'ex_coda']:
        coda = text[-1]
        text = text[:-1]
    return coda, text

def remove_tone(text: str) -> str:
    """Remove tone mark.

    Args:
        text: Thai syllable.

    Returns:
        Thai syllable without tone mark.
    """
    for char in TONE_MARK:
        text = text.replace(char, '')
    return text

def analyze_onset(matched_text: str, vowel_form: str) -> Tuple[
        str, bool, str, str]:
    """Find onset.

    Args:
        matched_text: Word without coda and silent_after
        vowel_form: Form of the vowel ex. -+า

    Returns:
        Onset, leading_h, onset_front, onset_back
    """
    vowel_form = vowel_form.replace('+', '') # + is tone mark place
    char_before, char_after = vowel_form.split('-') # - is onset place
    pattern = f'(?<={char_before})[ก-ฮ]+(?={char_after})(?!์)'
    onset_back = re.search(re.compile(pattern), matched_text)[0]
    onset_back, leading_h = analyze_leading_h(onset_back)

    # check for onset before vowel such as ส from สเต็ก
    onset_front = ''
    if char_before:
        pattern_front = f'^[ก-ฮ]+(?={char_before})'
        found_onset = re.search(re.compile(pattern_front), matched_text)
        if found_onset:
            onset_front = found_onset[0]
    onset = ''.join([onset_front, onset_back])
    return onset, leading_h, onset_front, onset_back

def analyze_leading_h(onset: str) -> Tuple[str, bool]:
    """Find and separate ห นำ from onset.

    Args:
        onset: Onset characters

    Returns:
        Onset (without leading h), If this word has ห นำ
    """
    leading_h: bool = False
    if ('ห' in onset
            and len(onset) > onset.index('ห')+1
            and CONSONANTS[onset[onset.index('ห')+1]]['class']
                == 'low_single'):
        leading_h = True
        onset = onset.replace('ห', '')
    return onset, leading_h

def find_onset_pref(onset_front: str, onset_back: str,
        tone_mark: str) -> Dict[str, bool]:
    """Find preference data for onset chars.

    This function is used when we have two onset chars (ห นำ
    not included).

    Args:
        onset_front: Onset before of front vowel such as ส in สเตย์
        onset_back: Onset after front vowel such as ต in สเตย์
        tone_mark: Tone mark of the word.

    Returns:
        Onset preference data.
    """
    clear_vowel = True
    clear_vowel_onset = 'not_true_cluster'
    clear_vowel_tone_mark = False
    if tone_mark:
        if len(onset_back) == 2:
            clear_vowel_tone_mark = False
        elif len(onset_front) == 1:
            clear_vowel_tone_mark = True
    if onset_front + onset_back in CLUSTERS:
        if len(onset_back) == 2:
            clear_vowel_onset = 'not_true_cluster'
        elif len(onset_front) == 1:
            clear_vowel_onset = 'all'
    else:
        if len(onset_back) == 2:
            clear_vowel = False
        elif len(onset_front) == 1:
            clear_vowel = True
    return {'clear_vowel': clear_vowel,
        'clear_vowel_onset': clear_vowel_onset,
        'clear_vowel_tone_mark': clear_vowel_tone_mark}

def analyze_silent_before(text: str) -> str:
    """Find silent_before. This assumes text doesn't have coda.

    Args:
        text: Syllable without coda.

    Returns:
        silent_before
    """
    silent_before: str = ''
    if text[-1] == DIACRITICS['kaaran']:
        silent_before = text[-2]
    return silent_before

def change_o_r(vowel: str, coda: str) -> Tuple[str, str]:
    """Change vowel for อร as it's usually pronounced with ออ not โอะ.

    Args:
        vowel: Vowel with อ as placeholder.
        coda: Syllable coda.

    Returns:
        New vowel and new vowel form
    """
    form: str = ''
    if vowel == 'โอะ' and coda == 'ร':
        vowel = 'ออ'
        form = '-+'
    return vowel, form

def save_h_thoo(tone: int, h_present: bool) -> bool:
    """If ห นำ+อักษรต่ำเดี่ยว+ไม้โท is used, save it to pref.

    Args:
        tone: Tone number.
        h_present: If ห นำ is here.

    Returns:
        low_single_h_thoo pref.
    """
    low_single_h_thoo: bool = False
    if tone == 2 and h_present:
        low_single_h_thoo = True
    return low_single_h_thoo


PREF = {
    'clear_vowel': (True, False),
    'clear_vowel_onset': ('not_true_cluster', 'all'),
    'clear_vowel_tone_mark': (True, False),
    'split_true_cluster': (True, False),
    'split_false_cluster': (True, False),
    'split_leading_con': (True, False),
    'obvious_low_singles': (True, False),
    'obvious_h_low_single': (True, False),
    'onset_style': ('plain', 'phinthu', 'yaamakkaan', 'kaaran'),
    'onset_style_apply': ('not_h', 'h'),
    'silent_before_style':
        ('plain', 'phinthu', 'yaamakkaan', 'kaaran', 'hide'),
    'coda_style': ('plain', 'phinthu', 'yaamakkaan', 'kaaran'),
    'silent_after_style':
        ('plain', 'phinthu', 'yaamakkaan', 'kaaran', 'hide'),
    'vowel_no_coda': ('pair', 'silent_after'),
    'vowel_coda_form': (),
    'vowel_length': ('input', 'short', 'long'),
    'vowel_pair_form': (),
    'low_single_h_thoo': (True, False)
}

_DEFAULT_PREF = {
    'clear_vowel': True,
    'clear_vowel_onset': 'not_true_cluster',
    'clear_vowel_tone_mark': False,
    'split_true_cluster': False,
    'split_false_cluster': False,
    'split_leading_con': False,
    'obvious_low_singles': True,
    'obvious_h_low_single': True,
    'onset_style': 'plain',
    'onset_style_apply': 'not_h',
    'silent_before_style': 'kaaran',
    'coda_style': 'plain',
    'silent_after_style': 'kaaran',
    'vowel_no_coda': 'pair',
    'vowel_coda_form': {},
    'vowel_length': 'input',
    'vowel_pair_form': {},
    'low_single_h_thoo': False
}


from typing import Dict, List, Tuple

from khanaa.thai_script import (CLUSTERS, CONSONANTS, TONE_NOT_AVAILABLE,
    VOWELS, TONES)

def find_vowel_pair(vowel: str) -> str:
    """Return vowel length pair."""
    pair = VOWELS[vowel]['pair']
    if not pair:
        pair = vowel
    return pair

def find_vowel_length(vowel: str) -> str:
    """Return vowel length ('short' or 'long')."""
    return VOWELS[vowel]['length']

def find_letter_list() -> Dict[str, List[str]]:
    """Return all consonants, vowels and true clusters."""
    result: Dict[str, List[str]] = {
        'consonants': list(CONSONANTS.keys()),
        'vowels': list(VOWELS.keys()),
        'true clusters': list(CLUSTERS)}
    return result

def letter_by_property(letter_dict: Dict[str, Dict[str, str]],
        property: str, property_name: str) -> List[str]:
    """Return list of letters with provided properties.
    
    Ex. letter_by_property(CONSONANTS, 'coda_class', 'alive')
    """
    return [letter for letter in letter_dict
        if letter_dict[letter][property] == property_name]

def check_checked(vowel: str, coda: str) -> bool:
    """Check if the word is checked (dead) or not.
    
    Checked word in Thai:
    - Has no coda, short length (and no hidden coda like ไอ)
    - Has dead class coda"""
    checked: bool = False
    if (not coda
            and VOWELS[vowel]['length'] == 'short'
            and not VOWELS[vowel]['sound_coda']
        ):
        checked = True
    elif (coda
            and CONSONANTS[coda]['coda_class'] == 'dead'
        ):
        checked = True
    return checked

def find_tone_data(onset: str, vowel: str,
        coda: str) -> Tuple[str, str, str]:
    """Return onset_class, alive_dead, length.
    
    onset accepts only one character."""
    onset_class: str = CONSONANTS[onset]['class']
    if onset_class in ['low_pair', 'low_single']:
        onset_class = 'low'
    alive_dead: str = 'dead' if check_checked(
        vowel, coda) else 'alive'
    length: str = VOWELS[vowel]['length']
    return onset_class, alive_dead, length

def find_tone_phrase(onset_class: str, alive_dead: str, length: str):
    """Return word detail in format that is used in TONES."""
    word_detail: str
    if onset_class in ['high', 'low'] and alive_dead == 'dead':
        word_detail = f'{onset_class} {length} {alive_dead}'
    else:
        word_detail = f'{onset_class} {alive_dead}'
    return word_detail

def find_tone(onset: str, vowel: str, coda: str = '',
        tone_marker: str = '', h_present: bool = False):
    """Find what tone (เสียงวรรณยุกต์) this syllable has.
    
    onset accepts only one character.
    tone_marker can be '' (not present), 'mai_eek', 'mai_thoo',
        'mai_trii', 'mai_jattawaa'
    h_present can be True (ห นำ is present) or False

    return -1 (not valid), 0 (สามัญ), 1 (เอก), 2 (โท), 3 (ตรี), 4 (จัตวา)
    """
    tone_num: int = -1
    onset_class, alive_dead, length = find_tone_data(onset, vowel, coda)
    if h_present and onset_class == 'low':
        onset_class = 'high'
    tone_info: str = find_tone_phrase(onset_class, alive_dead, length)

    for tone in TONES:
        if (tone_info in TONES[tone]
                and not (tone in TONE_NOT_AVAILABLE
                    and tone_info in TONE_NOT_AVAILABLE[tone])
                and TONES[tone][tone_info] == ['', tone_marker]):
            tone_num = tone
            break

    return tone_num


from typing import Any, List

from khanaa.combination import Combination
from khanaa.thai_script import DIACRITICS

class Kham:
    """Spell Thai word from information provided.
    
    Attributes:
        form (str): Spelled form of the word.
        data (dict[str, Any]): All information about this word.
        onset (str): พยัญชนะต้น
        onset_index (int): Index of onset used when calculating tone.
        onset_main (str): Main onset used when calculating tone.
        onset_class (str): high, mid, low_pair, low_single
        vowel (str): สระ
        vowel_length (str): short, long
        silent_before (str): ตัวการันต์ก่อนตัวสะกด
        coda (str): ตัวสะกด
        coda_class (str): dead (ก, บ, ด) or alive (the rest).
        silent_after (str): ตัวการันต์หลังตัวสะกด
        tone (int): เสียงวรรณยุกต์
        is_possible_tone (bool): Check if the input tone is possible
            with this word.
        tone_realized (int): Tone in spelled word, different from tone
            if tone is -1 or not possible with the word.
        use_leading_h (bool): Check if ห นำ is used.
        use_pair_onset (bool): Check if onset is changed to
            its pair onset.
        tone_mark (str): รูปวรรณยุกต์
        is_checked (bool): Check if the word is checked (คำตาย).
    
    Methods:
        ipa (str): Return basic IPA pronunciation of the word.
        is_donee_end (bool): Check if onset of the following word
            can be interpreted as this word's coda.
        is_donor_end (bool): Check if coda of this word can be
            interpreted as the following word onset.
        is_donor_start (bool): Check if onset of this word can be
            interpreted as coda of the preceding word.
        all_tone (list[str]): Return 0-4 tone version of the word.
        homophone (list[str]): Return a list of homophone
    """

    def __init__(
            self,
            onset: str,
            vowel: str,
            silent_before: str = '',
            coda: str = '',
            silent_after: str = '',
            tone: int = -1,
            **pref: Any) -> None:
        """Init Kham.

        None of the keyword arguments here are required.

        Find all available Thai consonants, vowels, and true clusters
        from find_letter_list()

        Examples:

            spell = Kham(onset='ข', vowel='เอีย', coda='น')

            result = spell.form

            result should be 'เขียน'

            spell = Kham('สม', 'อุ', '', 'ท', 'ร', -1,
                silent_after_style='plain')

            result = spell.form

            result should be 'สมุทร'

        Args:
            onset: Initial consonant(s), can be one or more.
                (พยัญชนะต้น)
            vowel: Vowel (สระ). Write the vowel with อ as
                a placeholder. If vowel has ตัวสะกด ย, ว or j, w coda,
                put them here together with vowel. Ex. อา, ไอ, อาว, อัย
            silent_before: Silent consonant(s) before coda.
                Defaults to ''.
            coda: Final consonant, excluding j, w (ตัวสะกด
                ไม่รวม ย, ว). Defaults to ''.
            silent_after: Silent consonant(s) after coda.
                Defaults to ''.
            tone: Tone (เสียงวรรณยุกต์). -1 not specified,
                0 mid สามัญ, 1 low เอก, 2 falling โท, 3 high ตรี,
                4 rising จัตวา. Defaults to -1.

        Keyword Args:
            clear_vowel (bool): Turn on/off option to put the first
                consonant in front of vowel to make the pronunciation
                less ambiguous.
                Option: False/off (ex. เชว), True/on (ex. ชเว).
                Default: True
            clear_vowel_onset (str): If clear_vowel is turned on,
                select onset cluster type that clear_vowel will be
                used with.
                Option: 'not_true_cluster' (true cluster คำควบกล้ำแท้
                won't be separated ex. เกวน ชเวน), 'all' (ex. กเวน ชเวน)
                Default: 'not_true_cluster'
            clear_vowel_tone_mark (bool): If clear_vowel is turned on,
                select if clean_vowel will be used with the word with
                tone marker or not.
                Option: False/not use (ex. คเว เคว่), True/use
                (ex. คเว คเว่).
                Default: False
            split_true_cluster (bool): Turn on/off option to treat
                true cluster onset as separate syllables. (เลือกว่า
                จะให้คำควบกล้ำแท้ออกเสียง/ผันวรรณยุกต์ควบกัน หรือว่า
                ออกเสียง/ผันวรรณยุกต์แบบเรียงพยางค์)
                Option: False/off (ex. กร+อุ+น+1 = กรุ่น = กฺรุ่น),
                True/on (ex. กร+อุ+น+1 = กหรุ่น = กะ-หฺรุ่น)
                Default: False
            split_false_cluster (bool): Turn on/off option to treat
                false cluster onset as separate syllables.
                Option: False/off (ex. ศร+อี = ศรี = สี,
                ศร+อา = ศรา = สา), True/on (ex. ศร+อี = ศรี = สะ-รี,
                ศร+อา = ศรา = สะ-รา)
                Default: False
            split_leading_con (bool): Turn on/off option to treat
                leading consonant cluster (อักษรนำ) as separate
                syllable.
                Option: False/off (ex. สล+อัว = สลัว = สะ-หฺลัว,
                สล+โอ+ว์ = สโลว์ = สะ-โหฺล), True/on (ex. สล+อัว = สลัว
                = สะ-ลัว, สล+โอ+ว์ = สโลว์ = สะ-โล
                Default: False
            obvious_low_singles (bool): Turn on/off option to make
                single low-single low onset cluster less ambiguous.
                Option: False/off (ex. หนวี่), True/on (ex. นหวี่).
                Default: True
            obvious_h_low_single (bool): Turn on/off option to make
                ฮ-single low onset cluster less ambiguous.
                Option: False/off (ex. หว่า), True/on (ex. ฮหว่า).
                Default: True
            onset_style (str): Select onset style.
                Option: 'plain' (no diacritic), 'phinthu' (ใส่พินทุ),
                'yaamakkaan' (ใส่ยามักการ), 'kaaran' (ใส่การันต์)
                Default: 'plain'
            onset_style_apply (str): Select character that onset style
                will be applied.
                Option: 'not_h' (ห นำ won't have diacritic), 'h' (ห นำ
                will have diacritic)
                Default: 'not_h'
            silent_before_style (str): Select style of the silent
                consonant that comes before the coda.
                Option: 'plain', 'phinthu', 'yaamakkaan', 'kaaran',
                'hide'
                Default: 'kaaran'
            coda_style (str): Select coda style.
                Option: 'plain', 'phinthu', 'yaamakkaan', 'kaaran'
                Default: 'plain'
            silent_after_style (str): Select style of the silent
                consonant that comes after the coda.
                Option: 'plain', 'phinthu', 'yaamakkaan', 'kaaran',
                'hide'
                Default: 'kaaran'
            vowel_no_coda (str): Select what we should do when vowels
                (เออะ, เอียะ, เอือะ, อัวะ) without its own coda form
                have coda.
                Option: 'pair' (use its pair form ex. เอียะ + น
                => เอียน), 'silent_after' (push the coda to
                silent_after ex. เอียะ + น => เอียะน์)
                Default: 'pair'
            vowel_coda_form (dict[str, str]): Select specific
                vowel form to use when coda is present. Put vowel in
                dict key and put vowel form in its value.
                In vowel form, - is the consonant position and + is
                the tone marker position.
                Example: vowel_coda_form={'เออ': 'เ-+อ'} (the result
                will be, for example, เดอน instead of the default เดิน)
            vowel_length (str): Select vowel length.
                Option: 'input' (same as input), 'short' (if long vowel
                is in the input, the result will be with
                a short vowel), 'long' (vice versa)
                Default: 'input'
            vowel_pair_form (dict[str, str]): Select specific
                vowel pair form to use when the length is specified.
                Example: vowel_length='short',
                vowel_pair_form={'อาย': 'อัย'}
                (อาย will be shorten to อัย instead of the default ไอ)
            low_single_h_thoo (bool): Turn on/off option to use
                leading h + mai thoo for word with low single onset and
                falling tone.
                Option: False/off (ex. ม+อะ+น+2=มั่น), True/on
                (ex. ม+อะ+น+2=หมั้น)
                Default: False
        """
        self._spell = Combination(onset, vowel, silent_before, coda,
            silent_after, tone, **pref)
        self.onset: str = self._spell._onset
        self.onset_index: int = self._spell._onset_index
        self.onset_main: str = self._spell._onset_main
        self.onset_class: str = self._spell._onset_class
        self.vowel: str = self._spell._vowel
        self.vowel_length: str = self._spell._vowel_length
        self.silent_before: str = self._spell._silent_before
        self.coda: str = self._spell._coda
        self.coda_class: str = self._spell._coda_class
        self.silent_after: str = self._spell._silent_after
        self.tone: int = self._spell._tone
        self.is_possible_tone: bool = self._spell._is_possible_tone
        self.tone_realized: int = self._spell._tone_realized
        self.use_leading_h: bool = self._spell._use_leading_h
        self.use_pair_onset: bool = self._spell._use_pair_onset
        self.tone_mark: str = self._spell._tone_mark
        self.is_checked: bool = self._spell._is_checked

        self.form = self._spell.form

    def __repr__(self) -> str:
        letters = [self.onset, self.vowel]
        if self.silent_before:
            letters.append(self.silent_before + DIACRITICS['kaaran'])
        if self.coda:
            letters.append(self.coda)
        if self.silent_after:
            letters.append(self.silent_after + DIACRITICS['kaaran'])

        letters.append(str(self.tone))
        letter_phrase = '+'.join(letters)

        return repr(' '.join([self.form, '=', letter_phrase]))

    def ipa(self) -> str:
        """Return basic IPA pronunciation of the spelled syllable.
        
        Ex. เขียน kʰ iaː n ˩˩˦, วลี w aʔ ˦˥ . l iː ˧

        Things not supported:
            - Stress: It depends on word level rather than syllable
            level. Also no vowel and tone reduction.
            - Irregular reading: ม+ไอ+3=ไม้=maj˦˥ (short vowel)
            instead of maːj˦˥ (long vowel) as in real world.
            - Yaamakkaan pronunciation
        
        Notes:
            - In the default setting, onsets being คำควบกล้ำแท้
            คำควบกล้ำไม่แท้ อักษรนำ are taken into account for tone
            calculation. So สล+อัว+-1=สลัว=saʔ˨˩.luaː˩˩˦ and
            สล+โอ+ว์+-1=สโลว์=saʔ˨˩.loː˩˩˦. But we can change this in
            split_leading_con so that สล+โอ+ว์+-1=สโลว์=saʔ˨˩.loː˧.
        """
        return self._spell.reading

    def is_donee_end(self) -> bool:
        """Return True if onset after this word can be seen as its coda.
        
        This function will return True if this word doesn't have coda,
        but if it is followed by word with more than one onset,
        the beginning onset of the following word can be interpreted
        as a coda of this word.
        
        Ex. ตา will return True. If ตา is followed by กลม, it will
        be ambiguous if the words ตากลม is ตา+กลม or ตาก+ลม.
        """
        return self._spell.is_donee_end

    def is_donor_end(self) -> bool:
        """Return True if coda can be seen as following word's onset.
        
        This function will return True if this word has coda (or
        some kind of unmarked silent_before or silent_after) and
        if this word is followed by another word, this coda can be
        interpreted as one of the onsets of the following word.

        Ex. ตาก will return True. If ตาก is followed by ลม, it will
        be ambiguous if the words ตากลม is ตา+กลม or ตาก+ลม.
        """
        return self._spell.is_donor_end

    def is_donor_start(self) -> bool:
        """Return True if onsets can be seen as precede word's coda.
        
        This function will return True if this word has more than
        one onsets, and if this word follows 'open end' word, the first
        onset of this word can be interpreted as coda of the word
        that comes before.

        Ex. กลม will return True. If the word ตา is followed by กลม,
        it will be ambiguous if the words ตากลม is ตา+กลม or ตาก+ลม.
        """
        return self._spell.is_donor_start
    
    def all_tone(self) -> List[str]:
        """Return 0-4 tone version of the word.
        
        Notes:
            - Now empty string will be the result if the word is not
            possible with that tone.
            Ex. ['', 'กะ', 'ก้ะ', 'ก๊ะ', 'ก๋ะ']
        """
        return self._spell.all_tone
    
    def homophone(self) -> List[str]:
        """Return a list of the word's homophone
        
        Notes:
            - The list also includes the original word
            - Silent characters are not included
        """
        return self._spell.homophone
    
    @property
    def data(self) -> dict[str, Any]:
        result = {attr: self.__dict__[attr] for attr in self.__dict__}
        result.pop("_spell")
        result.update({
            'ipa': self.ipa(),
            'is_donee_end': self.is_donee_end(),
            'is_donor_end': self.is_donor_end(),
            'is_donor_start': self.is_donor_start(),
            'all_tone': self.all_tone(),
            'homophone': self.homophone(),
        })
        return result


import unittest
from khanaa import Kham

GENERAL = [
    ({'onset': 'ค', 'vowel': 'โอะ', 'coda': 'น'},
        ['คญ', 'คณ', 'คน', 'คร', 'คล', 'คฬ', 'ฅญ', 'ฅณ', 'ฅน', 'ฅร', 'ฅล',
            'ฅฬ', 'ฆญ', 'ฆณ', 'ฆน', 'ฆร', 'ฆล', 'ฆฬ']),
    ({'onset': 'กว', 'vowel': 'อา', 'coda': 'ง', 'tone': 2},
        ['กว้าง']),
    ({'onset': 'สว', 'vowel': 'อิ', 'coda': 'ต', 'silent_after': 'ช'},
        ['สวิจ', 'สวิช', 'สวิซ', 'สวิฌ', 'สวิฎ', 'สวิฏ', 'สวิฐ', 'สวิฑ', 'สวิฒ', 'สวิด',
            'สวิต', 'สวิถ', 'สวิท', 'สวิธ', 'สวิศ', 'สวิษ', 'สวิส'])
]

class TestHomophone(unittest.TestCase):

    def test_general(self):
        for case in GENERAL:
            kham = Kham(**case[0])
            result = kham.homophone()
            self.assertEqual(result, case[1])

if __name__ == '__main__':
    unittest.main()


import unittest
from khanaa import Kham

GENERAL = [
    ({'onset': 'ส', 'vowel': 'เอีย', 'coda': 'ง', 'tone': -1},
    'เสียง'),
    ({'onset': 'ต', 'vowel': 'อะ', 'coda': 'ง', 'tone': 2},
    'ตั้ง'),
    ({'onset': 'ย', 'vowel': 'อะ', 'coda': 'ก', 'silent_after': 'ษ'},
    'ยักษ์'),
    ({'onset': 'บ', 'vowel': 'เออ', 'silent_before': 'ร', 'coda': 'น'},
    'เบิร์น'),
    ({'onset': 'พฤ', 'vowel': 'อ', 'coda': 'ก', 'silent_after': 'ษ'},
    'พฤกษ์'),
    ({'onset': 'ม', 'vowel': 'อิ', 'coda': 'น', 'silent_after': 'สก'},
    'มินสก์'),
    ({'onset': 'สก', 'vowel': 'เอะ', 'coda': 'ต'},
    'สเก็ต'),
    ({'onset': 'คว', 'vowel': 'เอ', 'coda': 'น'},
    'เควน'),
    ({'onset': 'สตร', 'vowel': 'เอ', 'coda': 'ส'},
    'สเตรส')
]

ONSET_TONE = [
    # alive mid, high, paired low, single low onsets
    ({'onset': 'ก', 'vowel': 'อา'},
    ['กา', 'ก่า', 'ก้า', 'ก๊า', 'ก๋า']),
    ({'onset': 'ข', 'vowel': 'อา'},
    ['คา', 'ข่า', 'ข้า', 'ค้า', 'ขา']),
    ({'onset': 'ค', 'vowel': 'อา'},
    ['คา', 'ข่า', 'ค่า', 'ค้า', 'ขา']),
    ({'onset': 'ง', 'vowel': 'อา'},
    ['งา', 'หง่า', 'ง่า', 'ง้า', 'หงา']),
    # dead mid, high, paired low, single low onsets (long)
    ({'onset': 'ก', 'vowel': 'อา', 'coda': 'บ'},
    ['', 'กาบ', 'ก้าบ', 'ก๊าบ', 'ก๋าบ']),
    ({'onset': 'ข', 'vowel': 'อา', 'coda': 'บ'},
    ['', 'ขาบ', 'ข้าบ', 'ค้าบ', 'ค๋าบ']),
    ({'onset': 'ค', 'vowel': 'อา', 'coda': 'บ'},
    ['', 'ขาบ', 'คาบ', 'ค้าบ', 'ค๋าบ']),
    ({'onset': 'ง', 'vowel': 'อา', 'coda': 'บ'},
    ['', 'หงาบ', 'งาบ', 'ง้าบ', 'ง๋าบ']),
    # dead mid, high, paired low, single low onsets (short)
    ({'onset': 'ก', 'vowel': 'อะ', 'coda': 'บ'},
    ['', 'กับ', 'กั้บ', 'กั๊บ', 'กั๋บ']),
    ({'onset': 'ข', 'vowel': 'อะ', 'coda': 'บ'},
    ['', 'ขับ', 'ขั้บ', 'คับ', 'คั๋บ']),
    ({'onset': 'ค', 'vowel': 'อะ', 'coda': 'บ'},
    ['', 'ขับ', 'คั่บ', 'คับ', 'คั๋บ']),
    ({'onset': 'ง', 'vowel': 'อะ', 'coda': 'บ'},
    ['', 'หงับ', 'งั่บ', 'งับ', 'งั๋บ'])]

ONSET_CLUSTER = [
    # mid & single low
    ({'onset': 'กว', 'vowel': 'อา'},
    ['กวา', 'กว่า', 'กว้า', 'กว๊า', 'กว๋า']),
    # pair low & single low
    ({'onset': 'ซย', 'vowel': 'อา'},
    ['ซยา', 'สย่า', 'ซย่า', 'ซย้า', 'สยา']),
    # high & high
    ({'onset': 'ผส', 'vowel': 'โอะ', 'coda': 'ม'},
    ['ผซม', 'ผส่ม', 'ผส้ม', 'ผซ้ม', 'ผสม']),
    # single low & single low
    ({'onset': 'ลว', 'vowel': 'อี'},
    ['ลวี', 'ลหวี่', 'ลวี่', 'ลวี้', 'ลหวี']),
    # three consonants
    ({'onset': 'สตร', 'vowel': 'อา'},
    ['สตรา', 'สตร่า', 'สตร้า', 'สตร๊า', 'สตร๋า'])
]

SETTING_STYLE = [
    ({'onset_style': 'kaaran'},
    {'onset': 'ทซ', 'vowel': 'อุ'},
    'ท์ซุ'),
    ({'onset_style': 'phinthu', 'silent_before_style': 'phinthu',
    'coda_style': 'phinthu', 'silent_after_style': 'phinthu'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'สฺเวิรฺลฺสฺ'),
    ({'onset_style': 'phinthu'},
    {'onset': 'ยว', 'vowel': 'อิ', 'coda': 'น', 'tone': 4},
    'ยฺหวิน'),
    ({'onset_style': 'yaamakkaan', 'silent_before_style': 'yaamakkaan',
    'coda_style': 'yaamakkaan', 'silent_after_style': 'yaamakkaan'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'ส๎เวิร๎ล๎ส๎'),
    ({'silent_before_style': 'plain',
    'silent_after_style': 'plain'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'สเวิรลส'),
    ({'silent_before_style': 'hide',
    'silent_after_style': 'hide'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'สเวิล'),
    ({'vowel_length': 'short'},
    {'onset': 'ม', 'vowel': 'อา', 'coda': 'น'},
    'มัน')]

SETTING_FORM = [
    ({'vowel_no_coda': 'pair'},
    {'onset': 'อ', 'vowel': 'เอียะ', 'coda': 'น'},
    'เอียน'),
    ({'vowel_no_coda': 'silent_after'},
    {'onset': 'อ', 'vowel': 'เอียะ', 'coda': 'น'},
    'เอียะน์'),
    ({'vowel_coda_form': {'เออ': 'เ-ิ+'}},
    {'onset': 'น', 'vowel': 'เออ', 'coda': 'ส'},
    'เนิส'),
    ({'vowel_coda_form': {'เออ': 'เ-+อ'}},
    {'onset': 'น', 'vowel': 'เออ', 'coda': 'ส'},
    'เนอส'),
    ({'vowel_length': 'long'},
    {'onset': 'ม', 'vowel': 'อะ', 'coda': 'น'},
    'มาน'),
    ({'vowel_length': 'short', 'vowel_pair_form': {'อาย': 'อัย'}},
    {'onset': 'อ', 'vowel': 'อาย'},
    'อัย'),
    ({'vowel_length': 'short', 'vowel_pair_form': {'อาย': 'ไอ'}},
    {'onset': 'อ', 'vowel': 'อาย'},
    'ไอ')
]

SETTING_PLACING = [
    ({'clear_vowel_onset': 'not_true_cluster'},
    {'onset': 'คว', 'vowel': 'เอ'},
    'เคว'),
    ({'clear_vowel_onset': 'all'},
    {'onset': 'คว', 'vowel': 'เอ'},
    'คเว'),
    ({'clear_vowel_onset': 'all', 'clear_vowel_tone_mark': False},
    {'onset': 'คว', 'vowel': 'เอ', 'tone': 2},
    'เคว่'),
    ({'clear_vowel_onset': 'all', 'clear_vowel_tone_mark': True},
    {'onset': 'คว', 'vowel': 'เอ', 'tone': 2},
    'คเว่')
]

SETTING_LS = [
    ({'obvious_low_singles': False},
    {'onset': 'ลว', 'vowel': 'อี'},
    ['ลวี', 'หลวี่', 'ลวี่', 'ลวี้', 'หลวี']),
    ({'obvious_low_singles': True},
    {'onset': 'ลว', 'vowel': 'อี'},
    ['ลวี', 'ลหวี่', 'ลวี่', 'ลวี้', 'ลหวี'])
]

SETTING_HLS = [
    ({'obvious_h_low_single': False},
    {'onset': 'ฮว', 'vowel': 'เอีย', 'coda': 'น'},
    ['ฮเวียน', 'เหวี่ยน', 'เฮวี่ยน', 'เฮวี้ยน', 'หเวียน']),
    ({'obvious_h_low_single': True},
    {'onset': 'ฮว', 'vowel': 'เอีย', 'coda': 'น'},
    ['ฮเวียน', 'ฮเหวี่ยน', 'เฮวี่ยน', 'เฮวี้ยน', 'ฮเหวียน'])
]

SETTING_SPLIT = [
    ({'split_true_cluster': False},
    {'onset': 'กร', 'vowel': 'อุ', 'coda': 'น', 'tone': 1},
    'กรุ่น'),
    ({'split_true_cluster': True},
    {'onset': 'กร', 'vowel': 'อุ', 'coda': 'น', 'tone': 1},
    'กหรุ่น'),
    ({'split_false_cluster': False},
    {'onset': 'ศร', 'vowel': 'อี', 'tone': 4},
    'ศรี'),
    ({'split_false_cluster': True},
    {'onset': 'ศร', 'vowel': 'อี', 'tone': 4},
    'ศหรี'),
    ({'split_leading_con': False},
    {'onset': 'สล', 'vowel': 'โอ', 'silent_before': 'ว', 'tone': 4},
    'สโลว์'),
    ({'split_leading_con': True},
    {'onset': 'สล', 'vowel': 'โอ', 'silent_before': 'ว', 'tone': 0},
    'สโลว์')
]

SETTING_H_THOO = [
    ({'low_single_h_thoo': False},
    {'onset': 'ม', 'vowel': 'อะ', 'coda': 'น', 'tone': 2},
    'มั่น'),
    ({'low_single_h_thoo': True},
    {'onset': 'ม', 'vowel': 'อะ', 'coda': 'น', 'tone': 2},
    'หมั้น')
]

TO_IPA = (
    ({},
    {'onset': 'ส', 'vowel': 'เอีย', 'coda': 'ง', 'tone': -1},
    's iaː ŋ ˩˩˦'),
    ({},
    {'onset': 'ต', 'vowel': 'อะ', 'coda': 'ง', 'tone': 2},
    't a ŋ ˥˩'),
    ({'split_leading_con': False},
    {'onset': 'สล', 'vowel': 'โอ', 'silent_before': 'ว', 'tone': 4},
    's aʔ ˨˩ . l oː ˩˩˦'),
    ({'split_leading_con': True},
    {'onset': 'สล', 'vowel': 'โอ', 'silent_before': 'ว', 'tone': 0},
    's aʔ ˨˩ . l oː ˧')
)

DONEE_END = [
    ({},
    {'onset': 'ต', 'vowel': 'อา'},
    True),
    ({},
    {'onset': 'ปล', 'vowel': 'อิว'},
    False),
    ({'silent_before_style': 'kaaran'},
    {'onset': 'อ', 'vowel': 'อา', 'silent_before': 'ร'},
    False),
    ({'silent_before_style': 'plain'},
    {'onset': 'อ', 'vowel': 'อา', 'silent_before': 'ร'},
    False)
]

DONOR_END = [
    ({},
    {'onset': 'ต', 'vowel': 'อา', 'coda': 'ก'},
    True),
    ({},
    {'onset': 'ค', 'vowel': 'อือ'},
    False),
    ({'silent_before_style': 'kaaran'},
    {'onset': 'อ', 'vowel': 'อา', 'silent_before': 'ร'},
    False),
    ({'silent_before_style': 'plain'},
    {'onset': 'อ', 'vowel': 'อา', 'silent_before': 'ร'},
    True)
]

DONOR_START = [
    ({},
    {'onset': 'กล', 'vowel': 'โอะ', 'coda': 'ม'},
    True),
    ({'clear_vowel': False},
    {'onset': 'คว', 'vowel': 'เอ'},
    False)
]

class TestSpellWord(unittest.TestCase):

    def test_general(self):
        for case in GENERAL:
            spell = Kham(**case[0])
            self.assertEqual(spell.form, case[1])

    def test_onset_tone(self):
        for case in ONSET_TONE:
            spell = Kham(**case[0])
            self.assertEqual(spell.all_tone(), case[1])

    def test_onset_cluster(self):
        for case in ONSET_CLUSTER:
            spell = Kham(**case[0])
            self.assertEqual(spell.all_tone(), case[1])

    def test_setting_style(self):
        for case in SETTING_STYLE:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.form, case[2])

    def test_setting_form(self):
        for case in SETTING_FORM:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.form, case[2])

    def test_setting_placing(self):
        for case in SETTING_PLACING:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.form, case[2])

    def test_setting_ls(self):
        for case in SETTING_LS:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.all_tone(), case[2])

    def test_setting_ls(self):
        for case in SETTING_HLS:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.all_tone(), case[2])

    def test_setting_split(self):
        for case in SETTING_SPLIT:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.form, case[2])

    def test_setting_h_thoo(self):
        for case in SETTING_H_THOO:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.form, case[2])

    def test_to_ipa(self):
        for case in TO_IPA:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.ipa(), case[2])

    def test_donee_end(self):
        for case in DONEE_END:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.is_donee_end(), case[2])

    def test_donor_end(self):
        for case in DONOR_END:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.is_donor_end(), case[2])

    def test_donor_start(self):
        for case in DONOR_START:
            spell = Kham(**case[0], **case[1])
            self.assertEqual(spell.is_donor_start(), case[2])

if __name__ == '__main__':
    unittest.main()


"""This module contains functions that are used for spelling word."""

from typing import Any, Dict
from khanaa.thai_script import CONSONANTS, DIACRITICS

def combine(word: Dict[str, Any], pref: Dict[str, Any]) -> str:
    """Main function for combining word data and spelling word.

    Args:
        word: Word data.
        pref: Preference data.

    Returns:
        Spelled form of the word.
    """
    index_after_vowel: int = find_index_after_vowel(word._onset,
        word._is_vowel_vague, word._is_low_single_vague,
        word._is_h_vague, word._use_leading_h)
    onset_convert: str = convert_onset(word._onset_main,
        word._use_pair_onset, word._use_leading_h)
    combined_onset_vowel: str = join_onset_vowel(index_after_vowel,
        word._use_leading_h, pref['onset_style'], pref['onset_style_apply'],
        word._vowel_form, word._onset, word._onset_index,
        onset_convert)
    combined_coda: str = combine_coda(word._silent_before,
        word._coda, word._silent_after, pref)
    combined: str = ''.join([combined_onset_vowel, combined_coda])
    combined = delete_taikhuu(combined, word._tone_mark)

    return combined.replace('+', word._tone_mark)

def find_index_after_vowel(onset: str, is_vowel_vague: bool,
        is_low_single_vague: bool, is_h_vague: bool,
        use_leading_h: bool) -> int:
    """Return index_after_vowel.
    
    index_after_vowel is the index of onset that
    should be after the vowel. Note that ห นำ is still not
    being calculated in this index yet.

    If the amount of onsets are more than two,
    just put the vowel in front of the penultimate onset.

    Then, for onsets with the amount of two,
    check for some Thai vowels that create ambiguity in
    pronunciation of the word with onset cluster.
    We might put these vowels in front of the last consonant
    to clarify the pronunciation.
    Ex. เชว, แชว, โชว > ชเว, ชแว, ชโว

    If the word is already mark for being ambiguous single low-
    single low cluster and has ห นำ, just put the vowel in front of the
    last consonant.

    Same case with ambiguous ฮ and single low cluster.

    Otherwise use general case: Every onset is after the vowel."""
    index_after_vowel: int = 0
    if len(onset) > 2:
        index_after_vowel = -2
    elif is_vowel_vague:
        index_after_vowel = -1
    elif is_low_single_vague and use_leading_h:
        index_after_vowel = -1
    elif is_h_vague and use_leading_h:
        index_after_vowel = -1
    
    return index_after_vowel

def convert_onset(onset_main, use_pair_onset, use_leading_h):
    """Convert onset to its pair or add ห นำ if it's needed."""
    onset_convert: str = onset_main
    if use_pair_onset:
        onset_convert = CONSONANTS[onset_main]['pair']
    elif use_leading_h:
        onset_convert = ''.join(['ห', onset_main])
    return onset_convert

def join_onset_vowel(index_after_vowel: int, use_leading_h: bool,
        onset_style: str, onset_style_apply: str, vowel_form: str,
        onset: str, onset_index: int, onset_main: str) -> str:
    """Join onsets and vowel according to the index provided."""
    combined_onset: str = join_onset(onset, onset_index, onset_main)
    index_after_vowel: int = find_new_index(use_leading_h, onset_index,
        index_after_vowel)

    before_vowel: str = combine_diacritic(onset_style,
        combined_onset[:index_after_vowel])
    after_vowel: str = combined_onset[index_after_vowel:]
    if len(after_vowel) > 1:
        diacritic_index: int = -1
        if (use_leading_h and onset_style_apply == 'not_h'):
            diacritic_index: int = -2
        after_vowel = ''.join([
            combine_diacritic(onset_style, after_vowel[:diacritic_index]),
            after_vowel[diacritic_index:]
        ])
    
    combined: str = ''.join([
        before_vowel,
        vowel_form.replace('-', after_vowel)
    ])
    return combined

def join_onset(onset: str, onset_index: int, onset_convert: str) -> str:
    """Join the form of main onset with other onsets."""
    prior: str = onset[:onset_index]
    latter: str = onset[onset_index+1:]
    if onset_index == -1:
        latter = ''
    combined_onset: str = ''.join([prior, onset_convert, latter])
    return combined_onset

def find_new_index(use_leading_h: bool, onset_index: int, index_after_vowel: int) -> int:
    """Shift index leftward if ห นำ is used."""
    if use_leading_h and onset_index >= index_after_vowel:
        index_after_vowel -= 1
    return index_after_vowel

def combine_diacritic(onset_style: str, content: str) -> str:
    """Combine diacritic to the onset.
    
    If option specifies that we should add diacritic to onset,
    add diacritic to every character in onset except the last one.
    """
    if onset_style in ['plain']:
        return content
    if onset_style in ['phinthu', 'yaamakkaan', 'kaaran']:
        diacritic = DIACRITICS[f"{onset_style}"]
        return add_diacritic(content, diacritic)
    raise ValueError('onset_style not recognized')

def add_diacritic(chars: str, diacritic: str) -> str:
    """Add diacritic to every character."""
    if chars:
        chars = ''.join([''.join([x, diacritic]) for x in chars])
    return chars

def add_diacritic_last(chars: str, diacritic: str) -> str:
    """Add diacritic to the last character."""
    if chars:
        chars = ''.join([chars, diacritic])
    return chars

def combine_coda(silent_before: str, coda: str, silent_after: str,
        pref: Dict[str, Any]) -> str:
    """Combine coda section: silent before, coda, silent after."""
    silent_before = combine_diacritic_coda(pref,
        silent_before, 'silent_before')
    coda = combine_diacritic_coda(pref, coda, 'coda')
    silent_after = combine_diacritic_coda(pref,
        silent_after, 'silent_after')

    combined_coda: str = ''.join([silent_before, coda, silent_after])
    return combined_coda

def combine_diacritic_coda(pref: Dict[str, Any], content: str,
        name: str) -> str:
    """Combine coda section with diacritic, depending on option."""
    style: str = f'{name}_style'
    if pref[style] in ['plain']:
        return content
    elif pref[style] in ['hide'] and name != 'coda':
        return ''
    elif pref[style] in [
            'phinthu', 'yaamakkaan', 'kaaran']:
        diacritic = DIACRITICS[f'{pref[style]}']
        if pref[style] in ['phinthu', 'yaamakkaan']:
            add_method = add_diacritic
        elif pref[style] in ['kaaran']:
            add_method = add_diacritic_last
        return add_method(content, diacritic)
    else:
        raise ValueError(f'{style} not recognized')

def delete_taikhuu(combined: str, tone_mark: str) -> str:
    """Delete mai taikhuu if there's also a tone marker."""
    mai_taikhuu: str = DIACRITICS['mai_taikhuu']
    if tone_mark and combined.find(mai_taikhuu) != -1:
        combined = combined.replace(mai_taikhuu, '')
    return combined


"""This module contains class that find word data."""

from typing import Any, Dict, List

from khanaa.setting import _DEFAULT_PREF
from khanaa.thai_script import (CLUSTERS, CONSONANTS, FALSE_CLUSTERS,
    LOW_SINGLE_ALT, TONE_MARKERS, TONE_NOT_AVAILABLE, TONES, VOWELS)
from khanaa.utils import (check_checked, find_tone, find_tone_phrase,
    find_vowel_length, find_vowel_pair)

class Word:
    """Find word/syllable data from input."""
    def __init__(
            self,
            onset: str,
            vowel: str,
            silent_before: str = '',
            coda: str = '',
            silent_after: str = '',
            tone: int = -1,
            **pref: Any) -> None:
        """See input information from Kham.
        
        Every attribute is usable and completely calculated except
        self._vowel_check and self._coda_check which are just
        calculation stage for self._vowel and self._coda.
        All attribute here won't be changed after it is defined and
        all calculation steps for them are in static methods began
        with _find + its name.
        """
        self.onset = onset
        self.vowel = vowel
        self.silent_before = silent_before
        self.coda = coda
        self.silent_after = silent_after
        self.tone = tone
        
        self.pref: Dict[str, Any] = {**_DEFAULT_PREF, **pref}

        self._onset: str = self._find_onset(self.onset)
        self._is_low_single_vague: bool = self._find_is_low_single_vague(
            self.pref['obvious_low_singles'], self._onset)
        self._is_h_vague: bool = self._find_is_h_vague(
            self.pref['obvious_h_low_single'], self._onset)
        self._onset_index: int = self._find_onset_index(
            self._is_low_single_vague, self._is_h_vague, self._onset,
            self.pref['split_true_cluster'], self.pref['split_false_cluster'],
            self.pref['split_leading_con'])
        self._onset_main: str = self._find_onset_main(self._onset,
            self._onset_index)
        self._onset_class: str = self._find_onset_class(self._onset_main)

        # _vowel_check and _coda_check are just calculation stages.
        self._vowel_check: str = self._find_vowel_check(self.vowel,
            self.pref['vowel_length'], self.pref['vowel_pair_form'])
        self._coda_check: str = self._find_coda_check(self._vowel_check,
            self.coda)
        self._is_vowel_empty_form: bool = self._find_is_vowel_empty_form(
            self.pref['vowel_coda_form'], self._vowel_check,
            self._coda_check)

        self._vowel: str = self._find_vowel(self._vowel_check,
            self._is_vowel_empty_form, self.pref['vowel_no_coda'])
        self._vowel_length: str = self._find_vowel_length(self._vowel)

        self._silent_before: str = self._find_silent_before(
            self.silent_before)

        self._coda: str = self._find_coda(self._coda_check,
            self._is_vowel_empty_form, self.pref['vowel_no_coda'])
        self._coda_class: str = self._find_coda_class(self._coda)
        
        self._vowel_form: str = self._find_vowel_form(self._vowel, self._coda,
            self._is_vowel_empty_form, self.pref['vowel_no_coda'],
            self.pref['vowel_coda_form'])
        self._is_checked: bool = self._find_is_checked(self._vowel,
            self._coda)

        self._silent_after: str = self._find_silent_after(self._vowel,
            self.coda, self.silent_after, self._is_vowel_empty_form,
            self.pref['vowel_no_coda'])
        
        self._tone: int = self._find_tone(self.tone)

        self._tone_phrase: str = self._find_tone_phrase(self._onset_class,
            self._is_checked, self._vowel_length)
        self._tone_detail: List[str] = self._find_tone_detail(self._tone,
            self._tone_phrase, self.pref['low_single_h_thoo'])
        self._is_possible_tone: bool = self._find_is_possible_tone(self._tone,
            self._tone_phrase)
        self._tone_realized: int = self._find_tone_realized(self._tone,
            self._onset_main, self._vowel, self._coda, self._is_possible_tone)
        self._use_pair_onset: bool = self._find_use_pair_onset(self._tone,
            self._tone_detail, self._onset_class)
        self._use_leading_h: bool = self._find_use_leading_h(self._tone,
            self._tone_detail, self._onset_class)
        self._tone_mark: str = self._find_tone_mark(self._tone,
            self._tone_detail)

        self._is_vowel_vague = self._find_is_vowel_vague(self._onset,
            self.pref['clear_vowel'], self.pref['clear_vowel_onset'],
            self.pref['clear_vowel_tone_mark'], self._tone_mark)

    @staticmethod
    def _find_onset(onset: str) -> str:
        return onset
    
    @staticmethod
    def _find_is_low_single_vague(obvious_low_singles: bool,
            onset: str) -> bool:
        """Check for low single-low single onset cluster ambiguity.
        
        If true, we'll use the latter consonant as the main consonant
        to determine tone.
        Ex. นวี นหวี่ นวี่ นวี้ นหวี
        instead of นวี หนวี่ นวี่ นวี้ หนวี
        """
        return (obvious_low_singles
            and len(onset) > 1
            and CONSONANTS[onset[-2]]['class'] == 'low_single'
            and CONSONANTS[onset[-1]]['class'] == 'low_single')

    @staticmethod
    def _find_is_h_vague(obvious_h_low_single: bool, onset: str) -> bool:
        """Check for ฮ and single low onset cluster ambiguity.
        
        If true, we'll use the latter consonant as the main consonant
        to determine tone. (ฮ is paired low and will have the same tone marker
        with single low vowel).
        Ex. ฮวา ฮหว่า ฮว่า ฮว้า ฮหวา
        instead of ฮวา หว่า ฮว่า ฮว้า หวา
        """
        return (obvious_h_low_single
            and len(onset) > 1
            and onset[-2] == 'ฮ'
            and CONSONANTS[onset[-1]]['class'] == 'low_single')

    @staticmethod
    def _find_onset_index(is_low_single_vague: bool,
            is_h_vague: bool, onset: str, split_true_cluster: bool,
            split_false_cluster: bool, split_leading_con: bool) -> int:
        """Select the index of main onset to determine tone from it.
        
        According to general rule of Thai onset cluster:

        - If the latter consonant is in single low class,
        tone of the word follows the prior consonant
        (อักษรควบแท้ อักษรควบไม่แท้ อักษรนำที่ตามด้วยอักษรต่ำเดี่ยว)
        - If not, tone follows the latter consonant
        (อักษรนำที่ไม่ได้ตามด้วยอักษรต่ำเดี่ยว)

        All of single low consonants are sonorants.
        """
        index: int = -1
        if is_low_single_vague or is_h_vague:
            pass
        elif split_true_cluster and onset[-2:] in CLUSTERS:
            pass
        elif split_false_cluster and onset[-2:] in FALSE_CLUSTERS:
            pass
        elif (len(onset) > 1
                and CONSONANTS[onset[-1]]['class'] == 'low_single'
                and not split_leading_con):
            index = -2
        return index

    @staticmethod
    def _find_onset_main(onset: str, onset_index: int) -> str:
        """Find main onset to be used when calculating tone."""
        return onset[onset_index]

    @staticmethod
    def _find_onset_class(onset_main: str) -> str:
        """หมวดหมู่ของพยัญชนะต้น"""
        return CONSONANTS[onset_main]['class']

    @staticmethod
    def _find_vowel_check(vowel_input: str, vowel_length_pref: str,
            vowel_pair_form_pref: Dict[str, str]) -> str:
        """Change length of input vowel according to the setting."""
        vowel: str = vowel_input
        if vowel_length_pref not in ['short', 'long']:
            pass
        elif vowel_length_pref != find_vowel_length(vowel_input):
            if vowel_input in vowel_pair_form_pref:
                vowel = vowel_pair_form_pref[vowel_input]
            else:
                vowel = find_vowel_pair(vowel_input)
        return vowel

    @staticmethod    
    def _find_is_vowel_empty_form(vowel_coda_form: Dict[str, str],
            vowel: str, coda: str) -> bool:
        """Check if vowel coda form is empty.
        
        Some vowels don't have coda form, so there are two choices:
        - Use coda form from its pair
        - Change coda to silent letter
        The choice is up to the setting.
        """
        if vowel_coda_form.get(vowel):
            return False
        elif (coda
                and not VOWELS[vowel]['form_with_coda']):
            return True
        else:
            return False

    @staticmethod
    def _find_vowel(vowel_input: str, is_vowel_empty_form: bool,
            vowel_no_coda_pref: str) -> str:
        """Find vowel.

        This step check if vowel coda form is empty and has to be
        change to its pair form according to the setting. The vowel
        should be already converted according to length from
        _find_vowel_check.
        """
        vowel: str = vowel_input
        if is_vowel_empty_form and vowel_no_coda_pref == 'pair':
            vowel = VOWELS[vowel]['pair']
        return vowel

    @staticmethod
    def _find_vowel_form(vowel: str, coda: str, is_vowel_empty_form: bool,
            vowel_no_coda_pref: str, vowel_coda_form: Dict[str, str]) -> str:
        """Find vowel form to use from VOWELS"""
        vowel_form: str = VOWELS[vowel]['form_no_coda']
        if coda:
            if vowel_coda_form.get(vowel):
                vowel_form = vowel_coda_form[vowel]
            elif (is_vowel_empty_form
                    and vowel_no_coda_pref == 'silent_after'):
                vowel_form = VOWELS[vowel]['form_no_coda']
            else:
                vowel_form = VOWELS[vowel]['form_with_coda']
        return vowel_form

    @staticmethod
    def _find_vowel_length(vowel: str) -> str:
        """Find ultimate vowel length."""
        return VOWELS[vowel]['length']

    @staticmethod
    def _find_silent_before(silent_before) -> str:
        return silent_before

    @staticmethod
    def _find_coda_check(vowel_check: str, coda_input: str) -> str:
        """Check vowel.
        
        If the vowel already coda, change coda to silent after.        
        Ex. อัย เอา already have j, w coda. They can't have coda.
        And change all but the first coda to silent after.
        """
        if VOWELS[vowel_check]['sound_coda']:
            return ''
        else:
            return coda_input[:1]

    @staticmethod
    def _find_coda(coda: str, is_vowel_empty_form: bool,
            vowel_no_coda_pref: str) -> str:
        """Return empty coda in case of vowel without coda form."""
        if (is_vowel_empty_form
                and vowel_no_coda_pref == 'silent_after'):
            return ''
        else:
            return coda

    @staticmethod
    def _find_coda_class(coda: str) -> str:
        """Return coda class (dead or alive)."""
        if coda:
            return CONSONANTS[coda]['coda_class']
        else:
            return ''

    @staticmethod
    def _find_is_checked(vowel: str, coda: str) -> bool:
        return check_checked(vowel, coda)

    @staticmethod
    def _find_silent_after(vowel: str, coda: str, silent_after: str,
            is_vowel_empty_form: bool, vowel_no_coda_pref: str) -> str:
        """Return silent after.
        
        If the vowel already has coda or coda should be changed to
        silent after because vowel coda is empty, index is zero.
        Otherwise index is one for multiple coda case.
        """
        index: int = 1
        if (VOWELS[vowel]['sound_coda']
                or (is_vowel_empty_form
                    and vowel_no_coda_pref == 'silent_after')):
            index = 0
        return ''.join([coda[index:], silent_after])

    @staticmethod
    def _find_tone(tone) -> int:
        return tone

    @staticmethod
    def _find_tone_realized(tone: int, onset_main: str, vowel: str,
            coda: str, is_possible_tone: bool) -> int:
        if tone == -1 or not is_possible_tone:
            return find_tone(onset_main, vowel, coda)
        else:
            return tone

    @staticmethod
    def _find_tone_phrase(onset_class: str, is_checked: bool,
            vowel_length) -> str:
        """Tone phrase is used when searching tone detail in TONES."""
        if onset_class in ['low_pair', 'low_single']:
            onset_class = 'low'
        alive_dead: str =  'dead' if is_checked else 'alive'
        return find_tone_phrase(onset_class,
            alive_dead, vowel_length)

    @staticmethod
    def _find_tone_detail(tone: int, tone_phrase: str,
            low_single_h_thoo: bool) -> List[str]:
        """Tone detail is whether onset has to be changed to pair
        and what tone mark it is.
        """
        if tone == -1:
            return ['', '']
        elif low_single_h_thoo and tone == 2:
            return LOW_SINGLE_ALT
        else:
            return TONES[tone][tone_phrase]
    
    @staticmethod
    def _find_use_pair_onset(tone: int, tone_detail: List[str],
            onset_class: str) -> bool:
        if tone == -1:
            return False
        else:
            return (tone_detail[0] == 'pair'
                and onset_class in ['high', 'low_pair'])

    @staticmethod
    def _find_use_leading_h(tone: int, tone_detail: List[str],
            onset_class: str) -> bool:
        """ห นำ"""
        if tone == -1:
            return False
        else:
            return (tone_detail[0] == 'pair'
                and onset_class == 'low_single')
    
    @staticmethod
    def _find_tone_mark(tone: int, tone_detail: List[str]) -> str:
        if tone == -1 or not tone_detail[1]:
            return ''
        else:
            return TONE_MARKERS[tone_detail[1]]
    
    @staticmethod
    def _find_is_possible_tone(tone: int, tone_phrase: str) -> bool:
        """As some words can't be pronounced with some tones
        such as 0 tone with checked syllable.
        """
        return not (tone in TONE_NOT_AVAILABLE
            and tone_phrase in TONE_NOT_AVAILABLE[tone])

    @staticmethod
    def _find_is_vowel_vague(onset: str, clear_vowel: bool,
            clear_vowel_onset: str, clear_vowel_tone_mark: bool,
            tone_mark: str) -> bool:
        """Check if we should put initial onset in front of vowel.
        
        As some Thai vowels create ambiguity in pronunciation
        of the word with onset cluster, we might put the prior onset
        in front of these vowels to clarify it.

        Ex. เชว, แชว, โชว > ชเว, ชแว, ชโว
        """
        return (clear_vowel
            and len(onset) == 2
            and (clear_vowel_onset == 'all'
                or (clear_vowel_onset == 'not_true_cluster'
                    and onset not in CLUSTERS))
            and (clear_vowel_tone_mark
                or (not clear_vowel_tone_mark
                    and not tone_mark)))


import unittest
from khanaa import SpellWord

GENERAL = [
    ({'onset': 'ส', 'vowel': 'เอีย', 'coda': 'ง', 'tone': -1},
    'เสียง'),
    ({'onset': 'ต', 'vowel': 'อะ', 'coda': 'ง', 'tone': 2},
    'ตั้ง'),
    ({'onset': 'ย', 'vowel': 'อะ', 'coda': 'ก', 'silent_after': 'ษ'},
    'ยักษ์'),
    ({'onset': 'บ', 'vowel': 'เออ', 'silent_before': 'ร', 'coda': 'น'},
    'เบิร์น'),
    ({'onset': 'พฤ', 'vowel': 'อ', 'coda': 'ก', 'silent_after': 'ษ'},
    'พฤกษ์'),
    ({'onset': 'ม', 'vowel': 'อิ', 'coda': 'น', 'silent_after': 'สก'},
    'มินสก์'),
    ({'onset': 'สก', 'vowel': 'เอะ', 'coda': 'ต'},
    'สเก็ต'),
    ({'onset': 'คว', 'vowel': 'เอ', 'coda': 'น'},
    'เควน'),
    ({'onset': 'สตร', 'vowel': 'เอ', 'coda': 'ส'},
    'สเตรส')
]

ONSET_TONE = [
    # alive mid, high, paired low, single low onsets
    ({'onset': 'ก', 'vowel': 'อา'},
    ['กา', 'ก่า', 'ก้า', 'ก๊า', 'ก๋า']),
    ({'onset': 'ข', 'vowel': 'อา'},
    ['คา', 'ข่า', 'ข้า', 'ค้า', 'ขา']),
    ({'onset': 'ค', 'vowel': 'อา'},
    ['คา', 'ข่า', 'ค่า', 'ค้า', 'ขา']),
    ({'onset': 'ง', 'vowel': 'อา'},
    ['งา', 'หง่า', 'ง่า', 'ง้า', 'หงา']),
    # dead mid, high, paired low, single low onsets (long)
    ({'onset': 'ก', 'vowel': 'อา', 'coda': 'บ'},
    ['กาบ', 'กาบ', 'ก้าบ', 'ก๊าบ', 'กาบ']),
    ({'onset': 'ข', 'vowel': 'อา', 'coda': 'บ'},
    ['ขาบ', 'ขาบ', 'ข้าบ', 'ค้าบ', 'ขาบ']),
    ({'onset': 'ค', 'vowel': 'อา', 'coda': 'บ'},
    ['คาบ', 'ขาบ', 'คาบ', 'ค้าบ', 'คาบ']),
    ({'onset': 'ง', 'vowel': 'อา', 'coda': 'บ'},
    ['งาบ', 'หงาบ', 'งาบ', 'ง้าบ', 'งาบ']),
    # dead mid, high, paired low, single low onsets (short)
    ({'onset': 'ก', 'vowel': 'อะ', 'coda': 'บ'},
    ['กับ', 'กับ', 'กั้บ', 'กั๊บ', 'กับ']),
    ({'onset': 'ข', 'vowel': 'อะ', 'coda': 'บ'},
    ['ขับ', 'ขับ', 'ขั้บ', 'คับ', 'ขับ']),
    ({'onset': 'ค', 'vowel': 'อะ', 'coda': 'บ'},
    ['คับ', 'ขับ', 'คั่บ', 'คับ', 'คับ']),
    ({'onset': 'ง', 'vowel': 'อะ', 'coda': 'บ'},
    ['งับ', 'หงับ', 'งั่บ', 'งับ', 'งับ'])]

ONSET_CLUSTER = [
    # mid & single low
    ({'onset': 'กว', 'vowel': 'อา'},
    ['กวา', 'กว่า', 'กว้า', 'กว๊า', 'กว๋า']),
    # pair low & single low
    ({'onset': 'ซย', 'vowel': 'อา'},
    ['ซยา', 'สย่า', 'ซย่า', 'ซย้า', 'สยา']),
    # high & high
    ({'onset': 'ผส', 'vowel': 'โอะ', 'coda': 'ม'},
    ['ผซม', 'ผส่ม', 'ผส้ม', 'ผซ้ม', 'ผสม']),
    # single low & single low
    ({'onset': 'ลว', 'vowel': 'อี'},
    ['ลวี', 'ลหวี่', 'ลวี่', 'ลวี้', 'ลหวี']),
    # three consonants
    ({'onset': 'สตร', 'vowel': 'อา'},
    ['สตรา', 'สตร่า', 'สตร้า', 'สตร๊า', 'สตร๋า'])
]

SETTING_STYLE = [
    ({'onset_style': 'kaaran'},
    {'onset': 'ทซ', 'vowel': 'อุ'},
    'ท์ซุ'),
    ({'onset_style': 'phinthu', 'silent_before_style': 'phinthu',
    'coda_style': 'phinthu', 'silent_after_style': 'phinthu'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'สฺเวิรฺลฺสฺ'),
    ({'onset_style': 'phinthu'},
    {'onset': 'ยว', 'vowel': 'อิ', 'coda': 'น', 'tone': 4},
    'ยฺหวิน'),
    ({'onset_style': 'yaamakkaan', 'silent_before_style': 'yaamakkaan',
    'coda_style': 'yaamakkaan', 'silent_after_style': 'yaamakkaan'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'ส๎เวิร๎ล๎ส๎'),
    ({'silent_before_style': 'plain',
    'silent_after_style': 'plain'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'สเวิรลส'),
    ({'silent_before_style': 'hide',
    'silent_after_style': 'hide'},
    {'onset': 'สว', 'vowel': 'เออ', 'silent_before': 'ร',
    'coda': 'ล', 'silent_after': 'ส'},
    'สเวิล'),
    ({'vowel_length': 'short'},
    {'onset': 'ม', 'vowel': 'อา', 'coda': 'น'},
    'มัน')]

SETTING_FORM = [
    ({'vowel_no_coda': 'pair'},
    {'onset': 'อ', 'vowel': 'เอียะ', 'coda': 'น'},
    'เอียน'),
    ({'vowel_no_coda': 'silent_after'},
    {'onset': 'อ', 'vowel': 'เอียะ', 'coda': 'น'},
    'เอียะน์'),
    ({'vowel_coda_form': {'เออ': 'เ-ิ+'}},
    {'onset': 'น', 'vowel': 'เออ', 'coda': 'ส'},
    'เนิส'),
    ({'vowel_coda_form': {'เออ': 'เ-+อ'}},
    {'onset': 'น', 'vowel': 'เออ', 'coda': 'ส'},
    'เนอส'),
    ({'vowel_length': 'long'},
    {'onset': 'ม', 'vowel': 'อะ', 'coda': 'น'},
    'มาน'),
    ({'vowel_length': 'short', 'vowel_pair_form': {'อาย': 'อัย'}},
    {'onset': 'อ', 'vowel': 'อาย'},
    'อัย'),
    ({'vowel_length': 'short', 'vowel_pair_form': {'อาย': 'ไอ'}},
    {'onset': 'อ', 'vowel': 'อาย'},
    'ไอ')
]

SETTING_PLACING = [
    ({'clear_vowel_onset': 'not_true_cluster'},
    {'onset': 'คว', 'vowel': 'เอ'},
    'เคว'),
    ({'clear_vowel_onset': 'all'},
    {'onset': 'คว', 'vowel': 'เอ'},
    'คเว'),
    ({'clear_vowel_onset': 'all', 'clear_vowel_tone_mark': False},
    {'onset': 'คว', 'vowel': 'เอ', 'tone': 2},
    'เคว่'),
    ({'clear_vowel_onset': 'all', 'clear_vowel_tone_mark': True},
    {'onset': 'คว', 'vowel': 'เอ', 'tone': 2},
    'คเว่')
]

SETTING_LS = [
    ({'obvious_low_singles': False},
    {'onset': 'ลว', 'vowel': 'อี'},
    ['ลวี', 'หลวี่', 'ลวี่', 'ลวี้', 'หลวี']),
    ({'obvious_low_singles': True},
    {'onset': 'ลว', 'vowel': 'อี'},
    ['ลวี', 'ลหวี่', 'ลวี่', 'ลวี้', 'ลหวี'])
]

SETTING_HLS = [
    ({'obvious_h_low_single': False},
    {'onset': 'ฮว', 'vowel': 'เอีย', 'coda': 'น'},
    ['ฮเวียน', 'เหวี่ยน', 'เฮวี่ยน', 'เฮวี้ยน', 'หเวียน']),
    ({'obvious_h_low_single': True},
    {'onset': 'ฮว', 'vowel': 'เอีย', 'coda': 'น'},
    ['ฮเวียน', 'ฮเหวี่ยน', 'เฮวี่ยน', 'เฮวี้ยน', 'ฮเหวียน'])
]

class TestSpellWord(unittest.TestCase):

    def test_general(self):
        spell = SpellWord()
        for case in GENERAL:
            self.assertEqual(spell.spell_out(**case[0]), case[1])

    def test_onset_tone(self):
        spell = SpellWord()
        for case in ONSET_TONE:
            self.assertEqual(spell.all_tone(**case[0]), case[1])

    def test_onset_cluster(self):
        spell = SpellWord()
        for case in ONSET_CLUSTER:
            self.assertEqual(spell.all_tone(**case[0]), case[1])

    def test_setting_style(self):
        for case in SETTING_STYLE:
            spell = SpellWord(**case[0])
            self.assertEqual(spell.spell_out(**case[1]), case[2])

    def test_setting_form(self):
        for case in SETTING_FORM:
            spell = SpellWord(**case[0])
            self.assertEqual(spell.spell_out(**case[1]), case[2])

    def test_setting_placing(self):
        for case in SETTING_PLACING:
            spell = SpellWord(**case[0])
            self.assertEqual(spell.spell_out(**case[1]), case[2])

    def test_setting_ls(self):
        for case in SETTING_LS:
            spell = SpellWord(**case[0])
            self.assertEqual(spell.all_tone(**case[1]), case[2])

    def test_setting_ls(self):
        for case in SETTING_HLS:
            spell = SpellWord(**case[0])
            self.assertEqual(spell.all_tone(**case[1]), case[2])

if __name__ == '__main__':
    unittest.main()


"""This module stores functions that check syllable's ambiguity."""

from khanaa.thai_script import CONSONANTS, DIACRITICS, VOWELS

def find_donee_end(vowel: str, silent_before: str, coda: str,
        silent_after: str, silent_before_style: str,
        silent_after_style: str) -> bool:
    """See details at Kham."""
    is_donee_end: bool = False
    if (VOWELS[vowel]['form_no_coda'] == VOWELS[vowel]['form_with_coda']
            and not coda
            and (not silent_before or silent_before_style == 'hide')
            and (not silent_after or silent_after_style == 'hide')
            and vowel not in ['อๅ']):
        is_donee_end = True
    return is_donee_end

def find_donor_end(vowel: str, silent_before: str, coda: str,
        silent_after: str, silent_before_style: str,
        silent_after_style: str, tone_mark: str,
        vowel_form: str) -> bool:
    """See details at Kham."""
    is_donor_end: bool = False
    if silent_after and silent_after_style == 'plain':
        is_donor_end = True
    elif silent_after and silent_after_style != 'hide':
        is_donor_end = False
    elif not coda and silent_before and silent_before_style == 'plain':
        is_donor_end = True
    elif coda and find_donor_end_coda(vowel, vowel_form, tone_mark):
        is_donor_end = True
    elif find_donor_end_jw(vowel, silent_before, silent_before_style,
            tone_mark):
        is_donor_end = True
    return is_donor_end

def find_donor_end_coda(vowel: str, vowel_form: str, tone_mark: str) -> bool:
    return (VOWELS[vowel]['form_no_coda'] == VOWELS[vowel]['form_with_coda']
        or (DIACRITICS['mai_taikhuu'] in VOWELS[vowel]['form_with_coda']
            and VOWELS[VOWELS[vowel]['pair']]['form_no_coda']
                == VOWELS[vowel]['form_with_coda'].replace(
                    DIACRITICS['mai_taikhuu'], '')
            and tone_mark)
        or vowel_form == '-+')

def find_donor_end_jw(vowel: str, silent_before: str,
        silent_before_style: str, tone_mark: str):
    return ((not silent_before or silent_before_style in ['plain', 'hide'])
        and VOWELS[vowel]['sound_coda'] in ['j', 'w']
        and vowel[-1] in ['ย', 'ว']
        and (DIACRITICS['mai_taikhuu'] not in vowel
            or tone_mark)
        and vowel[:-1].replace(DIACRITICS['mai_taikhuu'], '')
            in VOWELS
        and (VOWELS[vowel[:-1].replace(
                DIACRITICS['mai_taikhuu'], '')]['form_no_coda']
            == VOWELS[vowel[:-1].replace(
                DIACRITICS['mai_taikhuu'], '')]['form_with_coda']
            or vowel[:-1].replace(DIACRITICS['mai_taikhuu'], '')
                == 'เออ'))

def find_donor_start(onset: str, form: str) -> bool:
    """See details at Kham."""
    is_donor_start: bool = False
    if (len(onset) > 1
            and form[0] in CONSONANTS
            and CONSONANTS[form[0]]['sound_coda']
            and form[0] not in ['ห', 'ฮ']):
        is_donor_start = True
    return is_donor_start


"""This module contains class used to derive word pronunciation."""

from typing import Any

from khanaa.thai_script import (CLUSTERS, CONSONANTS, FALSE_CLUSTERS, VOWELS,
    TONE_IPA)
from khanaa.thai_spelling import find_vowel_length
from khanaa.utils import find_tone

class ThaiToIPA:
    """Convert Thai spelling data to IPA."""
    def __init__(
            self,
            onset: str,
            vowel: str,
            silent_before: str = '',
            coda: str = '',
            silent_after: str = '',
            tone: int = -1,
            **pref: Any) -> None:
        """Input should be already cleaned from Word and tone should
        be realized tone instead of the input tone.
        """
        self.onset = onset
        self.vowel = vowel
        self.silent_before = silent_before
        self.coda = coda
        self.silent_after = silent_after
        self.tone = tone
        
        self.pref = pref

    def convert(self) -> str:
        """Return basic IPA pronunciation from data provided."""
        joined_vowel = ''.join([self.vowel_ipa, self.length_ipa])
        return ' '.join(filter(None, [self.onset_ipa, joined_vowel,
            self.vowel_coda_ipa, self.coda_ipa, self.tone_ipa]))
    
    @property
    def is_true_cluster(self) -> bool:
        true_cluster = False
        if (self.pref['split_true_cluster'] == False
                and len(self.onset) > 1
                and self.onset[-2:] in CLUSTERS):
            true_cluster = True
        return true_cluster
    
    @property
    def is_false_cluster(self) -> bool:
        false_cluster = False
        if (self.pref['split_false_cluster'] == False
                and len(self.onset) > 1
                and self.onset[-2:] in FALSE_CLUSTERS):
            false_cluster = True
        return false_cluster

    @property
    def onset_ipa(self) -> str:
        onset_ipa_chars = []
        convert_onset = self.onset
        if self.is_false_cluster:
            # because false clusters have ซ sound
            convert_onset = convert_onset[:-2] + 'ซ'
        reversed_onset = convert_onset[::-1]
        for index, char in enumerate(reversed_onset):
            if not (index == 0
                    or (index == 1 and self.is_true_cluster)):
                # insert them backward because we're in reverse mode
                onset_ipa_chars.extend(['.',
                    self.find_tone_onset(char), 'aʔ'])
            onset_ipa_chars.append(CONSONANTS[char]['sound_onset'])
        return ' '.join(reversed(onset_ipa_chars))
    
    @staticmethod
    def find_tone_onset(onset_char) -> str:
        """Find IPA tone for syllabic onset character."""
        tone_num = find_tone(onset_char, 'อะ', '', '', False)
        return TONE_IPA[tone_num]

    @property
    def vowel_ipa(self) -> str:
        return VOWELS[self.vowel]['sound_vowel']
    
    @property
    def vowel_coda_ipa(self) -> str:
        return VOWELS[self.vowel]['sound_coda']
    
    @property
    def length_ipa(self) -> str:
        """Return IPA symbol according to vowel length."""
        length_char = ''
        if (find_vowel_length(self.vowel) == 'long'
                or self.pref['vowel_length'] == 'long'):
            length_char = 'ː'
        elif ((not self.coda and not self.vowel_coda_ipa)
                or self.pref['vowel_length'] == 'short'):
            length_char = 'ʔ'
        return length_char
    
    @property
    def coda_ipa(self) -> str:
        coda_char = ''
        if self.coda:
            coda_char = CONSONANTS[self.coda]['sound_coda']
        return coda_char
    
    @property
    def tone_ipa(self) -> str:
        tone_char = ''
        if self.tone in range(5):
            tone_char = TONE_IPA[self.tone]
        return tone_char


from itertools import product
from khanaa.thai_script import CONSONANTS, VOWELS

def find_same_sound_consonant(consonant: str, type: str) -> list[str]:
    """Find a list of consonants that have same sound as the input

    Args:
        consonant: a consonant character
        type: "onset" or "coda"

    Returns:
        List of consonants (including the input)
    """
    result = []
    if len(consonant) != 1:
        return result
    name: str = f"sound_{type}"
    sound: str = CONSONANTS.get(consonant)[name]
    for con in CONSONANTS:
        if CONSONANTS[con][name] == sound:
            result.append(con)
    
    return result

def find_same_sound_vowel(vowel: str) -> list[str]:
    """Find a list of vowels that have same sound as the input

    Args:
        vowel: a vowel with default อ

    Returns:
        List of vowels (including the input)
    """
    result = []
    data: str = VOWELS.get(vowel)
    for v in VOWELS:
        v_data = VOWELS[v]
        same_length: bool = data["length"] == v_data["length"]
        same_sound_vowel: bool = data["sound_vowel"] == v_data["sound_vowel"]
        same_sound_coda: bool = data["sound_coda"] == v_data["sound_coda"]
        if same_length and same_sound_vowel and same_sound_coda:
            result.append(v)
    
    return result

def find_homophone_product(onset: str, vowel: str, coda: str):
    """Find a product of inputs' homophone

    Args:
        onset: initial consonant (one or more character)
        vowel: a vowel with default อ
        coda: final consonant

    Returns:
        Product of possible tuples (onset, vowel, coda)
    """
    if len(onset) == 1:
        onsets = find_same_sound_consonant(onset, "onset")
    else:
        onsets = [onset]
    vowels = find_same_sound_vowel(vowel)
    codas = find_same_sound_consonant(coda, "coda")
    if not coda:
        codas.append("")
        if "อรร" in vowels:
            vowels.remove("อรร")
    elif "อรร" in vowels and "ร" in codas:
        codas.remove("ร")
        codas.append("")
    
    possibles = list(product(onsets, vowels, codas))
    
    if "อำ" in vowels:
        a_vowels = find_same_sound_vowel("อะ")
        m_codas = find_same_sound_consonant("ม", "coda")
        new_possibles = list(product(onsets, a_vowels, m_codas))
        possibles.extend(new_possibles)

    elif "อะ" in vowels and coda == "ม":
        possibles.append((onset, "อำ", ""))
    
    return possibles